#include "image.h"

/* Internal functions */

int nc_fgets(char *s, int m, FILE *f);

Image *CreateImage(int ncols, int nrows)
{
  Image *img=NULL;
  int i;

  img = (Image *) calloc(1,sizeof(Image));
  if (img == NULL){
    Error(MSG1,"CreateImage");
  }

  img->val   = AllocIntArray(nrows*ncols);
  img->tbrow = AllocIntArray(nrows);

  img->tbrow[0]=0;
  for (i=1; i < nrows; i++)
    img->tbrow[i]=img->tbrow[i-1]+ncols;
  img->ncols = ncols;
  img->nrows = nrows;
 
 return(img);
}

void DestroyImage(Image **img)
{
  Image *aux;

  aux = *img;
  if(aux != NULL){
    if (aux->val != NULL)   free(aux->val); 
    if (aux->tbrow != NULL) free(aux->tbrow);
    free(aux);    
    *img = NULL;
  }
}

Image *ReadImage(char *filename)
{
  FILE *fp=NULL;
  unsigned char *value=NULL;
  char type[10];
  int  i,ncols,nrows,n;
  Image *img=NULL;
  char z[256];

  fp = fopen(filename,"rb");
  if (fp == NULL){
    fprintf(stderr,"Cannot open %s\n",filename);
    exit(-1);
  }
  fscanf(fp,"%s\n",type);
  if((strcmp(type,"P5")==0)){
    nc_fgets(z,255,fp);
    sscanf(z,"%d %d\n",&ncols,&nrows);
    n = ncols*nrows;
    nc_fgets(z,255,fp);
    sscanf(z,"%d\n",&i);
    value = (unsigned char *)calloc(n,sizeof(unsigned char));
    if (value != NULL){
      fread(value,sizeof(unsigned char),n,fp);
    }else{
      fprintf(stderr,"Insufficient memory in ReadImage\n");
      exit(-1);
    }
    fclose(fp);
    img = CreateImage(ncols,nrows);
    for (i=0; i < n; i++)
      img->val[i]=(int)value[i];
    free(value);
  }else{
    if((strcmp(type,"P2")==0)){
      nc_fgets(z,255,fp);
      sscanf(z,"%d %d\n",&ncols,&nrows);
      n = ncols*nrows;
      nc_fgets(z,255,fp);
      sscanf(z,"%d\n",&i);
      img = CreateImage(ncols,nrows);
      for (i=0; i < n; i++)
	fscanf(fp,"%d",&img->val[i]);
      fclose(fp);
    }else{
      fprintf(stderr,"Input image must be P2 or P5\n");
      exit(-1);
    }
  }

  return(img);
}

void WriteImage(Image *img,char *filename)
{
  FILE *fp;
  int i, n, Imax;

  fp = fopen(filename,"wb");
  if (fp == NULL){
    fprintf(stderr,"Cannot open %s\n",filename);
    exit(-1);
  }
  n    = img->ncols*img->nrows;
  if ((Imax=MaximumValue(img))==INT_MAX){
    Warning("Image with infinity values","WriteImage");
    Imax = INT_MIN;
    for (i=0; i < n; i++) 
      if ((img->val[i] > Imax)&&(img->val[i]!=INT_MAX))
	Imax = img->val[i];
    fprintf(fp,"P2\n");
    fprintf(fp,"%d %d\n",img->ncols,img->nrows);
    fprintf(fp,"%d\n",Imax+1);
  } else {
    fprintf(fp,"P2\n");
    fprintf(fp,"%d %d\n",img->ncols,img->nrows);
    if (Imax==0) Imax++;
    fprintf(fp,"%d\n",Imax);
  }
 
  for (i=0; i < n; i++) {
    if (img->val[i]==INT_MAX)
      fprintf(fp,"%d ",Imax+1);
    else
      fprintf(fp,"%d ",img->val[i]);
    if (((i+1)%17) == 0)
      fprintf(fp,"\n");
  }

  fclose(fp);

}

int MinimumValue(Image *img)
{
  int i,min,n;

  n = img->ncols*img->nrows;
  min = img->val[0];
  for (i=1; i < n; i++)
    if (img->val[i] < min)
      min = img->val[i];

  return(min);
}

int MaximumValue(Image *img)
{
  unsigned int i, n, r;
  int max;

  max = img->val[0];
  n = img->ncols*img->nrows - 1;
  r = n%4;
  n -= r;
  for (i=1; i < n; i+=4) {
    if (img->val[i] > max)
      max = img->val[i];
    if (img->val[i+1] > max)
      max = img->val[i+1];
    if (img->val[i+2] > max)
      max = img->val[i+2];
    if (img->val[i+3] > max)
      max = img->val[i+3];
  }
  while (r != 0) {
    if (img->val[i+r-1] > max)
      max = img->val[i+r-1];
    --r;
  }

  return(max);
}

bool ValidPixel(Image *img, int x, int y)
{
  if ((x >= 0)&&(x < img->ncols)&&
      (y >= 0)&&(y < img->nrows))
    return(true);
  else
    return(false);
}

Image *CopyImage(Image *img)
{
  Image *imgc;

  imgc = CreateImage(img->ncols,img->nrows);
  memcpy(imgc->val,img->val,img->ncols*img->nrows*sizeof(int));
  
  return(imgc);
}


Image *Sum(Image *img1, Image *img2)
{
  Image *sum;
  int p,n;

  sum = CopyImage(img1);
  n = img1->ncols*img1->nrows;
  for (p=0; p < n; p++)
    sum->val[p] += img2->val[p];
  return(sum);
}

Image *CreateCos(int ncols, int nrows, float wx, float wy, float phix, float phiy)
{
  int x, y, p;
  Image *cosimg = NULL;
  
  cosimg = CreateImage(ncols, nrows);
  
  /* We are adding a DC level of 127 to the amplitude of the
     sine. This will create an extra impulse at the center of the
     magnitude spectrum. */
  
  for (x = 0; x < ncols; x++) 
    for ( y = 0; y < nrows; y++) {
      p = x + cosimg->tbrow[y];
      cosimg->val[p] = (int) (127.0*(cos(2*PI*wx*(float)(x - phix)/ncols + 2*PI*wy*(float)(y - phiy)/nrows))+127.0+0.5);
    }
  
  return cosimg;
}

Image  *CreateSquare(int xo, int yo, int xf, int yf, int ncols, int nrows, int val)
{
  Image *img=CreateImage(MAX(ncols,xf-xo+1),MAX(nrows,yf-yo+1));
  int x,y;

  for (y=yo; y < yf; y++) 
    for (x=xo; x < xf; x++) {
	img->val[x + img->tbrow[y]]=val;      
    }  
  return(img);
}

Image  *CreateIntArray(char *data)
{
  int ncols,nrows,i,n;
  Image *img;

  sscanf(data,"%d",&ncols);
  data=strchr(data,',')+1;
  sscanf(data,"%d",&nrows);
  data=strchr(data,',')+1;

  img = CreateImage(ncols, nrows);
  n   = ncols*nrows;
  for (i=0;i<n;i++) {
    sscanf(data,"%d",&img->val[i]);
    data=strchr(data,',')+1;
  }
  return(img);
}

/* pula comentarios com # */
int nc_fgets(char *s, int m, FILE *f) {
  while(fgets(s,m,f)!=NULL)
    if (s[0]!='#') return 1;
  return 0;
}

/********************* Questo 1 ************************/

Image *AddZeros(Image *img, int Nx, int Ny)
{
  Image *zimg=NULL;

  return(zimg);
}

Image *RemZeros(Image *img, int Nx, int Ny)
{
  Image *zimg=NULL;

  return(zimg);
}
